"use strict";

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } return _assertThisInitialized(self); }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

var EventEmitter = require("events");

var http = require("http");

var _require = require("../functions"),
    formatError = _require.formatError;

var _require2 = require("../constants"),
    ERR_CODES = _require2.ERR_CODES,
    ERR_MSGS = _require2.ERR_MSGS;

var _require3 = require("../buffer"),
    MessageBuffer = _require3.MessageBuffer;
/**
 * @class Client
 * @extends require('events').EventEmitter
 * @param {Object} [server] object with host and port of server
 * @param {Object} [options]
 * @param {Number} [options.version=2] JSON-RPC version to use (1|2)
 * @param {String} [options.delimiter="\n"] delimiter to use for requests
 * @param {Number} [options.timeout=30] timeout for request
 * @return {Client}
 */


var Client =
/*#__PURE__*/
function (_EventEmitter) {
  _inherits(Client, _EventEmitter);

  function Client(options) {
    var _this;

    _classCallCheck(this, Client);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(Client).call(this));

    if (!(_assertThisInitialized(_this) instanceof Client)) {
      return _possibleConstructorReturn(_this, new Client(options));
    }

    var defaults = {
      host: "127.0.0.1",
      port: "8100",
      version: "2.0",
      delimiter: "\n",
      timeout: 30,
      retries: 2
    };
    _this.client = undefined;
    _this.message_id = 1;
    _this.serving_message_id = 1;
    _this.pendingCalls = {};
    _this.pendingBatches = {};
    _this.attached = false;
    _this.responseQueue = {};
    _this.options = _objectSpread({}, defaults, options || {});
    _this.options.timeout = _this.options.timeout * 1000;
    _this.messageBuffer = new MessageBuffer(_this.options.delimiter);
    var _this$options = _this.options,
        host = _this$options.host,
        port = _this$options.port;
    _this.server = {
      host: host,
      port: port
    };
    return _this;
  } // http client doesnt need this method


  _createClass(Client, [{
    key: "connect",
    value: function connect() {
      throw new Error("function must be overwritten in subclass");
    }
  }, {
    key: "end",
    value: function end() {
      var _this2 = this;

      return new Promise(function (resolve, reject) {
        _this2.client.end(function (error) {
          if (error) {
            reject(error);
          }

          resolve();
        });
      });
    }
  }, {
    key: "request",
    value: function request() {
      throw new Error("function must be overwritten in subclass");
    }
  }, {
    key: "batch",
    value: function batch() {
      throw new Error("function must be overwritten in subclass");
    }
  }, {
    key: "notify",
    value: function notify() {
      throw new Error("function must be overwritten in subsclass");
    }
  }, {
    key: "subscribe",
    value: function subscribe() {
      throw new Error("function must be overwritten in subsclass");
    }
  }, {
    key: "handleResponse",
    value: function handleResponse(id) {
      try {
        var response = this.responseQueue[id];

        if (this.writer instanceof http.IncomingMessage) {
          // want to allow users to access the headers, status code etc.
          response = _objectSpread({
            body: this.responseQueue[id]
          }, this.writer);
        }

        this.pendingCalls[id].resolve(response);
        delete this.responseQueue[id];
      } catch (e) {
        if (e instanceof TypeError) {
          // probably a parse error, which might not have an id
          process.stdout.write("Message has no outstanding calls: ".concat(JSON.stringify(e), "\n"));
        }
      }
    }
  }, {
    key: "verifyData",
    value: function verifyData(chunk) {
      var _this3 = this;

      try {
        // will throw an error if not valid json
        var message = JSON.parse(chunk);

        if (Array.isArray(message)) {
          // possible batch request
          message.forEach(function (res) {
            if (res && res.method && !res.id) {
              _this3.emit(res.method, res);
            }
          });
          this.emit("batchResponse", message);
        } else if (!(message === Object(message))) {
          // error out if it cant be parsed
          var error = formatError({
            jsonrpc: this.options.version,
            delimiter: this.options.delimiter,
            id: null,
            code: ERR_CODES.parseError,
            message: ERR_MSGS.parseError
          });
          throw new Error(error);
        } else if (!message.id) {
          // no id, so assume notification
          // special case http response since it cant get notifications
          // this is not in spec at all
          if (this.writer instanceof http.IncomingMessage) {
            var _error = formatError({
              jsonrpc: this.options.version,
              delimiter: this.options.delimiter,
              id: this.serving_message_id,
              code: ERR_CODES.parseError,
              message: ERR_MSGS.parseError
            });

            throw new Error(_error);
          }

          this.emit(message.method, message);
        } else if (message.error) {
          // got an error back so reject the message
          var _error2 = formatError({
            jsonrpc: this.options.version,
            delimiter: this.options.delimiter,
            id: message.id,
            code: message.error.code,
            message: message.error.message
          });

          throw new Error(_error2);
        } else if (!message.method) {
          // no method, so assume response
          this.serving_message_id = message.id;
          this.responseQueue[this.serving_message_id] = message;
          this.handleResponse(this.serving_message_id);
        } else {
          var _error3 = formatError({
            jsonrpc: this.options.version,
            delimiter: this.options.delimiter,
            id: null,
            code: ERR_CODES.unknown,
            message: ERR_MSGS.unknown
          });

          throw new Error(_error3);
        }
      } catch (e) {
        if (e instanceof SyntaxError) {
          var _error4 = formatError({
            jsonrpc: this.options.version,
            delimiter: this.options.delimiter,
            id: this.serving_message_id,
            code: ERR_CODES.parseError,
            message: "Unable to parse message: '".concat(chunk, "'")
          });

          throw new Error(_error4);
        } else {
          throw e;
        }
      }
    }
  }, {
    key: "listen",
    value: function listen() {
      var _this4 = this;

      this.writer.on("data", function (data) {
        if (_this4.writer instanceof http.IncomingMessage) {
          _this4.writer.on("end", function () {
            // may still want to know the data if there are errors
            _this4.writer.response = data.toString();

            _this4.handleData(data);
          });
        } else {
          _this4.handleData(data);
        }
      });
      this.client.on("close", function () {
        _this4.attached = false;

        _this4.client.removeAllListeners();

        _this4.emit("serverDisconnected");
      });
    }
  }, {
    key: "handleData",
    value: function handleData(data) {
      this.messageBuffer.push(data);

      while (!this.messageBuffer.isFinished()) {
        var message = this.messageBuffer.handleData();

        try {
          this.verifyData(message);
        } catch (e) {
          this.handleError(e);
        }
      }
    }
  }, {
    key: "serverDisconnected",
    value: function serverDisconnected(cb) {
      this.on("serverDisconnected", function () {
        return cb();
      });
    }
  }, {
    key: "handleError",
    value: function handleError(error) {
      var err;

      try {
        err = JSON.parse(error.message);
      } catch (e) {
        err = JSON.parse(formatError({
          jsonrpc: this.options.version,
          delimiter: this.options.delimiter,
          id: null,
          code: ERR_CODES.unknown,
          message: JSON.stringify(error, Object.getOwnPropertyNames(error))
        }));
      }

      var response = err;

      if (this.writer instanceof http.IncomingMessage) {
        // want to allow users to access the headers, status code etc.
        response = _objectSpread({
          body: err
        }, this.writer);
      }

      try {
        this.pendingCalls[err.id].reject(response);
      } catch (e) {
        if (e instanceof TypeError) {
          // probably a parse error, which might not have an id
          process.stdout.write("Message has no outstanding calls: ".concat(JSON.stringify(response), "\n"));
        }
      }
    }
  }]);

  return Client;
}(EventEmitter);

module.exports = Client;
/**
 * TCP client constructor
 * @type ClientTcp
 * @static
 */

Client.tcp = require("./tcp");
/**
 * HTTP client constructor
 * @type ClientHTTP
 * @static
 */

Client.http = require("./http");
/**
 * WebSoket client constructor
 * @type ClientWS
 * @static
 */

Client.ws = require("./ws");