"use strict";

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } return _assertThisInitialized(self); }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

var WebSocket = require("ws");

var Client = require(".");

var _require = require("../functions"),
    formatRequest = _require.formatRequest,
    formatError = _require.formatError;

var _require2 = require("../constants"),
    ERR_CODES = _require2.ERR_CODES,
    ERR_MSGS = _require2.ERR_MSGS;

var _require3 = require("../buffer"),
    MessageBuffer = _require3.MessageBuffer;

var WSClient =
/*#__PURE__*/
function (_Client) {
  _inherits(WSClient, _Client);

  function WSClient(options) {
    var _this;

    _classCallCheck(this, WSClient);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(WSClient).call(this));

    if (!(_assertThisInitialized(_this) instanceof WSClient)) {
      return _possibleConstructorReturn(_this, new WSClient(options));
    }

    var defaults = {
      url: "ws://127.0.0.1:8100",
      version: "2.0",
      delimiter: "\n",
      timeout: 30,
      retries: 2
    };
    _this.message_id = 1;
    _this.serving_message_id = 1;
    _this.pendingCalls = {};
    _this.pendingBatches = {};
    _this.attached = false;
    _this.responseQueue = {};
    _this.options = _objectSpread({}, defaults, options || {});
    _this.options.timeout = _this.options.timeout * 1000;
    _this.messageBuffer = new MessageBuffer(_this.options.delimiter);
    var retries = _this.options.retries;
    _this.remainingRetries = retries;
    return _this;
  }

  _createClass(WSClient, [{
    key: "connect",
    value: function connect() {
      var _this2 = this;

      return new Promise(function (resolve, reject) {
        var _this2$options = _this2.options,
            url = _this2$options.url,
            perMessageDeflate = _this2$options.perMessageDeflate;
        _this2.client = new WebSocket(url, perMessageDeflate);

        _this2.close();

        _this2.listen();

        _this2.client.onopen = function (event) {
          resolve(event);
        };

        _this2.client.onerror = function (error) {
          reject(error);
        };
      });
    }
  }, {
    key: "close",
    value: function close() {
      var _this3 = this;

      this.client.onclose = function () {
        if (_this3.remainingRetries) {
          _this3.remainingRetries -= 1;
          process.stdout.write("Connection failed. ".concat(_this3.remainingRetries, " attempts left.\n"));
          setTimeout(function () {
            _this3.connect()["catch"](function () {});
          }, _this3.options.timeout);
        }
      };
    }
  }, {
    key: "listen",
    value: function listen() {
      var _this4 = this;

      this.client.onmessage = function (message) {
        _this4.handleData(message.data);
      };
    }
  }, {
    key: "request",
    value: function request() {
      var _this5 = this;

      return {
        message: function message(method, params) {
          var id = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : true;
          var request = formatRequest({
            method: method,
            params: params,
            id: id ? _this5.message_id : undefined,
            options: _this5.options
          });

          if (id) {
            _this5.message_id += 1;
          }

          return request;
        },
        send: function send(method, params) {
          return new Promise(function (resolve, reject) {
            var requestId = _this5.message_id;
            _this5.pendingCalls[requestId] = {
              resolve: resolve,
              reject: reject
            };

            try {
              _this5.client.send(_this5.request().message(method, params));
            } catch (e) {
              reject(e);
            }

            setTimeout(function () {
              try {
                var error = JSON.parse(formatError({
                  jsonrpc: _this5.options.version,
                  delimiter: _this5.options.delimiter,
                  id: null,
                  code: ERR_CODES.timeout,
                  message: ERR_MSGS.timeout
                }));

                _this5.pendingCalls[requestId].reject(error);

                delete _this5.pendingCalls[requestId];
              } catch (e) {
                if (e instanceof TypeError) {
                  process.stdout.write("Message has no outstanding calls: ".concat(JSON.stringify(e), "\n"));
                }
              }
            }, _this5.options.timeout);
          });
        },
        notify: function notify(method, params) {
          var request = formatRequest({
            method: method,
            params: params,
            options: _this5.options
          });
          return new Promise(function (resolve, reject) {
            try {
              _this5.client.send(request);

              resolve(request);

              _this5.client.onerror = function (error) {
                reject(error);
              };
            } catch (e) {
              reject(e);
            }
          });
        }
      };
    }
  }, {
    key: "batch",
    value: function batch(requests) {
      var _this6 = this;

      /**
       * should receive a list of request objects
       * [client.request.message(), client.request.message()]
       * send a single request with that, server should handle it
       *
       * We want to store the IDs for the requests in the batch in an array
       * Use this to reference the batch response
       * The spec has no explaination on how to do that, so this is the solution
       */
      return new Promise(function (resolve, reject) {
        var batchIds = [];
        var batchRequests = [];
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
          for (var _iterator = requests[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            var _request = _step.value;
            var json = JSON.parse(_request);
            batchRequests.push(json);

            if (json.id) {
              batchIds.push(json.id);
            }
          }
        } catch (err) {
          _didIteratorError = true;
          _iteratorError = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion && _iterator["return"] != null) {
              _iterator["return"]();
            }
          } finally {
            if (_didIteratorError) {
              throw _iteratorError;
            }
          }
        }

        _this6.pendingBatches[String(batchIds)] = {
          resolve: resolve,
          reject: reject
        };
        var request = JSON.stringify(batchRequests);

        try {
          _this6.client.send(request + _this6.options.delimiter);
        } catch (e) {
          // this.client is probably undefined
          reject(e.message);
        }

        setTimeout(function () {
          try {
            var error = JSON.parse(formatError({
              jsonrpc: _this6.options.version,
              delimiter: _this6.options.delimiter,
              id: null,
              code: ERR_CODES.timeout,
              message: ERR_MSGS.timeout
            }));

            _this6.pendingBatches[String(batchIds)].reject(error);

            delete _this6.pendingBatches[String(batchIds)];
          } catch (e) {
            if (e instanceof TypeError) {
              process.stdout.write("Message has no outstanding calls: ".concat(JSON.stringify(e), "\n"));
            }
          }
        }, _this6.options.timeout);

        _this6.on("batchResponse", function (batch) {
          var batchResponseIds = [];
          batch.forEach(function (message) {
            if (message.id) {
              batchResponseIds.push(message.id);
            }
          });

          if (batchResponseIds.length === 0) {
            resolve([]);
          }

          var _loop = function _loop() {
            var ids = _Object$keys[_i];
            var arrays = [JSON.parse("[".concat(ids, "]")), batchResponseIds];
            var difference = arrays.reduce(function (a, b) {
              return a.filter(function (c) {
                return !b.includes(c);
              });
            });

            if (difference.length === 0) {
              batch.forEach(function (message) {
                if (message.error) {
                  // reject the whole message if there are any errors
                  try {
                    _this6.pendingBatches[ids].reject(batch);

                    delete _this6.pendingBatches[ids];
                  } catch (e) {
                    if (e instanceof TypeError) {// no outstanding calls
                    }
                  }
                }
              });

              try {
                _this6.pendingBatches[ids].resolve(batch);

                delete _this6.pendingBatches[ids];
              } catch (e) {
                if (e instanceof TypeError) {// no outstanding calls
                }
              }
            }
          };

          for (var _i = 0, _Object$keys = Object.keys(_this6.pendingBatches); _i < _Object$keys.length; _i++) {
            _loop();
          }
        });
      });
    }
    /**
     * @params {String} [method] method to subscribe to
     * @params {Function} [cb] callback function to invoke on notify
     */

  }, {
    key: "subscribe",
    value: function subscribe(method, cb) {
      if (method === "batchResponse") {
        throw new Error("\"batchResponse\" is a reserved event name");
      }

      this.on(method, cb);
    }
    /**
     * @params {String} [method] method to unsubscribe from
     * @params {Function} [cb] name of function to remove
     */

  }, {
    key: "unsubscribe",
    value: function unsubscribe(method, cb) {
      if (method === "batchResponse") {
        throw new Error("\"batchResponse\" is a reserved event name");
      }

      this.removeListener(method, cb);
    }
    /**
     * @params {String} [method] method to unsubscribe all listeners from
     */

  }, {
    key: "unsubscribeAll",
    value: function unsubscribeAll(method) {
      if (method === "batchResponse") {
        throw new Error("\"batchResponse\" is a reserved event name");
      }

      this.removeAllListeners([method]);
    }
  }]);

  return WSClient;
}(Client);

module.exports = WSClient;