"use strict";

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } return _assertThisInitialized(self); }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

var EventEmitter = require("events");

var _require = require("../functions"),
    formatResponse = _require.formatResponse,
    formatError = _require.formatError,
    BatchRequest = _require.BatchRequest;

var _require2 = require("../constants"),
    ERR_CODES = _require2.ERR_CODES,
    ERR_MSGS = _require2.ERR_MSGS;
/**
 * @class Server
 * @extends require('events').EventEmitter
 * @param {Object} [host] host IP to connect with
 * @param {Object} [host] host port to connect with
 * @param {Object} [options]
 * @param {Number} [options.version=2] JSON-RPC version to use (1|2)
 * @param {String} [options.delimiter="\n"] delimiter to use for requests
 * @param {Boolean} [options.exlusive=false] disallow port sharing
 * @return {Client}
 */


var Server =
/*#__PURE__*/
function (_EventEmitter) {
  _inherits(Server, _EventEmitter);

  function Server(options) {
    var _this;

    _classCallCheck(this, Server);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(Server).call(this));

    if (!(_assertThisInitialized(_this) instanceof Server)) {
      return _possibleConstructorReturn(_this, new Server(options));
    }

    var defaults = {
      host: "127.0.0.1",
      port: 8100,
      exclusive: false,
      version: "2.0",
      delimiter: "\n"
    };
    _this.options = _objectSpread({}, defaults, options || {});
    _this.methods = {};
    _this.listening = false;
    return _this;
  }

  _createClass(Server, [{
    key: "listen",
    value: function listen() {
      var _this2 = this;

      return new Promise(function (resolve, reject) {
        if (_this2.listening) {
          // not having this caused MaxEventListeners error
          reject(Error("server already listening"));
        }

        var _this2$options = _this2.options,
            host = _this2$options.host,
            port = _this2$options.port,
            exclusive = _this2$options.exclusive;

        _this2.server.listen({
          host: host,
          port: port,
          exclusive: exclusive
        });

        _this2.server.on("listening", function () {
          _this2.listening = true;

          try {
            _this2.handleData();

            _this2.handleError();

            resolve({
              host: _this2.server.address().address,
              port: _this2.server.address().port
            });
          } catch (e) {
            _this2.listening = false;
            reject(e);
          }
        });

        _this2.server.on("error", function (error) {
          _this2.listening = false;
          reject(error);
        });

        _this2.server.on("close", function () {
          _this2.listening = false;
        });
      });
    }
  }, {
    key: "close",
    value: function close() {
      var _this3 = this;

      this.listening = false;
      return new Promise(function (resolve, reject) {
        _this3.server.close(function (error) {
          if (error) {
            reject(error);
          }

          resolve();
        });
      });
    }
  }, {
    key: "method",
    value: function method(name, cb) {
      // add the method and its associated callback to the object
      this.methods[name] = cb;
    }
  }, {
    key: "onNotify",
    value: function onNotify(method, cb) {
      if (method === "clientConnected" || method === "clientDisconnected") {
        throw new Error("reserved event name ".concat(method));
      }

      this.on(method, cb);
    }
  }, {
    key: "removeOnNotify",
    value: function removeOnNotify(method, cb) {
      if (method === "clientConnected" || method === "clientDisconnected") {
        throw new Error("reserved event name ".concat(method));
      }

      this.removeListener(method, cb);
    }
  }, {
    key: "removeAllOnNotify",
    value: function removeAllOnNotify(method) {
      if (method === "clientConnected" || method === "clientDisconnected") {
        throw new Error("reserved event name ".concat(method));
      }

      this.removeAllListeners([method]);
    }
  }, {
    key: "handleData",
    value: function handleData() {
      throw new Error("function must be overwritten in subsclass");
    }
  }, {
    key: "handleBatchRequest",
    value: function handleBatchRequest(requests) {
      var _this4 = this;

      var batchRequests = requests.map(function (request) {
        try {
          var message = _this4.validateMessage(request);

          if (message.notification) {
            _this4.emit(message.notification.method, message.notification);

            return;
          }

          return _this4.getResult(message).then(function (result) {
            return JSON.parse(result);
          })["catch"](function (error) {
            throw error;
          });
        } catch (e) {
          return Promise.reject(e);
        }
      }).filter(function (el) {
        return el != null;
      });

      if (batchRequests.length === 0) {
        return Promise.resolve({
          empty: true
        });
      }

      return Promise.all(batchRequests.map(function (promise) {
        return promise["catch"](function (error) {
          return JSON.parse(error.message);
        });
      }));
    }
  }, {
    key: "validateRequest",
    value: function validateRequest(request) {
      try {
        var message = JSON.parse(request);
        return message;
      } catch (e) {
        throw new Error(formatError({
          jsonrpc: this.options.version,
          id: null,
          code: ERR_CODES.parseError,
          message: ERR_MSGS.parseError,
          delimiter: this.options.delimiter
        }));
      }
    }
  }, {
    key: "validateMessage",
    value: function validateMessage(message) {
      if (Array.isArray(message)) {
        // possible batch request
        if (message.length === 0) {
          var error = formatError({
            jsonrpc: this.options.version,
            id: null,
            code: ERR_CODES.invalidRequest,
            message: ERR_MSGS.invalidRequest,
            delimiter: this.options.delimiter
          });
          throw new Error(error);
        }

        throw new BatchRequest(undefined, message);
      }

      if (!(message === Object(message))) {
        throw new Error(formatError({
          jsonrpc: this.options.version,
          id: null,
          code: ERR_CODES.invalidRequest,
          message: ERR_MSGS.invalidRequest,
          delimiter: this.options.delimiter
        }));
      }

      if (!(typeof message.method === "string")) {
        throw new Error(formatError({
          jsonrpc: message.jsonrpc,
          id: message.id,
          code: ERR_CODES.invalidRequest,
          message: ERR_MSGS.invalidRequest,
          delimiter: this.options.delimiter
        }));
      }

      if (!message.id) {
        // no id, so assume notification
        return {
          notification: message
        };
      }

      if (message.jsonrpc) {
        if (this.options.version !== "2.0") {
          throw new Error(formatError({
            id: message.id,
            code: ERR_CODES.invalidRequest,
            message: ERR_MSGS.invalidRequest,
            delimiter: this.options.delimiter
          }));
        }
      }

      if (!this.methods[message.method]) {
        throw new Error(formatError({
          jsonrpc: message.jsonrpc,
          id: message.id,
          code: ERR_CODES.methodNotFound,
          message: ERR_MSGS.methodNotFound,
          delimiter: this.options.delimiter
        }));
      }

      if (message.params && !Array.isArray(message.params) && !(message.params === Object(message.params))) {
        throw new Error(formatError({
          jsonrpc: message.jsonrpc,
          id: message.id,
          code: ERR_CODES.invalidParams,
          message: ERR_MSGS.invalidParams,
          delimiter: this.options.delimiter
        }));
      } // data looks good


      return message;
    }
  }, {
    key: "handleValidation",
    value: function handleValidation(chunk) {
      var _this5 = this;

      return new Promise(function (resolve, reject) {
        try {
          var result = _this5.validateRequest(chunk);

          var message = _this5.validateMessage(result);

          resolve(message);
        } catch (e) {
          if (e instanceof BatchRequest) {
            _this5.handleBatchRequest(e.request).then(function (result) {
              resolve({
                batch: result
              });
            })["catch"](function (error) {
              reject(error);
            });
          } else {
            reject(e);
          }
        }
      });
    }
  }, {
    key: "getResult",
    value: function getResult(message) {
      var _this6 = this;

      // function needs to be async since the method can be a promise
      return new Promise(function (resolve, reject) {
        var params = message.params;
        var error = formatError({
          jsonrpc: message.jsonrpc,
          id: message.id,
          code: ERR_CODES.unknown,
          message: ERR_MSGS.unknown,
          delimiter: _this6.options.delimiter
        });

        try {
          var result = params ? _this6.methods[message.method](params) : _this6.methods[message.method]();

          if (result && (typeof result.then === "function" || result instanceof Promise)) {
            Promise.all([result]).then(function (results) {
              resolve(formatResponse({
                jsonrpc: message.jsonrpc,
                id: message.id,
                result: results || 0,
                delimiter: _this6.options.delimiter
              }));
            })["catch"](function (resError) {
              error = formatError({
                jsonrpc: message.jsonrpc,
                id: message.id,
                code: ERR_CODES.internal,
                message: "".concat(JSON.stringify(resError.message || resError)),
                delimiter: _this6.options.delimiter
              });
              reject(error);
            });
          } else {
            resolve(formatResponse({
              jsonrpc: message.jsonrpc,
              id: message.id,
              result: result || 0,
              delimiter: _this6.options.delimiter
            }));
          }
        } catch (e) {
          if (e instanceof TypeError) {
            error = formatError({
              jsonrpc: message.jsonrpc,
              id: message.id,
              code: ERR_CODES.invalidParams,
              message: ERR_MSGS.invalidParams,
              delimiter: _this6.options.delimiter
            });
          }

          reject(error);
        }
      });
    }
  }, {
    key: "clientConnected",
    value: function clientConnected() {
      throw new Error("function must be overwritten in subclass");
    }
  }, {
    key: "clientDisconnected",
    value: function clientDisconnected() {
      throw new Error("function must be overwritten in subsclass");
    }
  }, {
    key: "handleError",
    value: function handleError() {
      var _this7 = this;

      this.on("error", function (error) {
        _this7.listening = false;
        throw error;
      });
    }
  }]);

  return Server;
}(EventEmitter);

module.exports = Server;
/**
 * HTTP server constructor
 * @type ServerHTTP
 * @static
 */

Server.http = require("./http");
/**
 * TCP server constructor
 * @type ServerTCP
 * @static
 */

Server.tcp = require("./tcp");
/**
 * WS server constructor
 * @type ServerWS
 * @static
 */

Server.ws = require("./ws");