"use strict";

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _slicedToArray(arr, i) { return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest(); }

function _nonIterableRest() { throw new TypeError("Invalid attempt to destructure non-iterable instance"); }

function _iterableToArrayLimit(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"] != null) _i["return"](); } finally { if (_d) throw _e; } } return _arr; }

function _arrayWithHoles(arr) { if (Array.isArray(arr)) return arr; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } return _assertThisInitialized(self); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

var net = require("net");

var Server = require(".");

var _require = require("../functions"),
    formatResponse = _require.formatResponse;

var _require2 = require("../ServerProtocol"),
    TCPServerProtocol = _require2.TCPServerProtocol;
/**
 * Constructor for Jsonic TCP client
 * @class TCPClient
 * @constructor
 * @extends Client
 * @param {Object} [options] optional settings for client
 * @return TCPClient
 */


var TCPServer =
/*#__PURE__*/
function (_Server) {
  _inherits(TCPServer, _Server);

  function TCPServer(options) {
    var _this;

    _classCallCheck(this, TCPServer);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(TCPServer).call(this, options));
    _this.connectedClients = [];

    _this.initServer();

    return _this;
  }

  _createClass(TCPServer, [{
    key: "initServer",
    value: function initServer() {
      this.server = new net.Server();
    }
  }, {
    key: "handleData",
    value: function handleData() {
      var _this2 = this;

      this.server.on("connection", function (client) {
        _this2.connectedClients.push(client);

        _this2.emit("clientConnected", client);

        var tcpServerProtocol = new TCPServerProtocol(_this2, client, _this2.options.delimiter);
        tcpServerProtocol.clientConnected();
      });
    }
  }, {
    key: "clientConnected",
    value: function clientConnected(cb) {
      this.on("clientConnected", function (client) {
        cb({
          host: client.remoteAddress,
          port: client.remotePort
        });
      });
    }
  }, {
    key: "clientDisconnected",
    value: function clientDisconnected(cb) {
      var _this3 = this;

      this.on("clientDisconnected", function (client) {
        var clientIndex = _this3.connectedClients.findIndex(function (c) {
          return client === c;
        });

        if (clientIndex === -1) {
          return cb("Unknown client ".concat(JSON.stringify(client)));
        }

        var _this3$connectedClien = _this3.connectedClients.splice(clientIndex, 1),
            _this3$connectedClien2 = _slicedToArray(_this3$connectedClien, 1),
            deletedClient = _this3$connectedClien2[0];

        return cb({
          host: deletedClient.remoteAddress,
          port: deletedClient.remotePort
        });
      });
    } // only available for TCP and ws server

  }, {
    key: "notify",
    value: function notify(notifications) {
      var _this4 = this;

      if (notifications.length === 0 || !Array.isArray(notifications)) {
        throw new Error("Invalid arguments");
      }

      var responses = notifications.map(function (_ref) {
        var _ref2 = _slicedToArray(_ref, 2),
            method = _ref2[0],
            params = _ref2[1];

        if (!method && !params) {
          throw new Error("Unable to generate a response object");
        }

        var response = _this4.options.version === "2.0" ? {
          jsonrpc: "2.0",
          method: method,
          params: params,
          delimiter: _this4.options.delimiter
        } : {
          method: method,
          params: params,
          delimiter: _this4.options.delimiter
        };
        return response;
      });

      if (responses.length === 0) {
        throw new Error("Unable to generate a response object");
      }

      var response;

      if (responses.length === 1) {
        response = formatResponse(responses[0]);
      } else {
        response = "[";
        responses.forEach(function (res, idx) {
          response += formatResponse(res);
          response += idx === responses.length - 1 ? "" : ",";
        });
        response += "]";
      }
      /**
       * Returns list of error objects if there was an error sending to any client
       */


      if (this.connectedClients.length === 0) {
        return [new Error("No clients connected")];
      }

      return this.connectedClients.map(function (client) {
        try {
          return client.write(JSON.stringify(JSON.parse(response)) + _this4.options.delimiter);
        } catch (e) {
          // possibly client disconnected
          return e;
        }
      });
    }
  }]);

  return TCPServer;
}(Server);

module.exports = TCPServer;